
import { localize } from '../../../../nls.js';
import { onUnexpectedError } from '../../../../base/common/errors.js';
import Severity$1 from '../../../../base/common/severity.js';
import { EXTENSION_IDENTIFIER_PATTERN } from '../../../../platform/extensionManagement/common/extensionManagement.js';
import { Extensions } from '../../../../platform/jsonschemas/common/jsonContributionRegistry.js';
import { Registry } from '../../../../platform/registry/common/platform.js';
import { ExtensionIdentifierSet, EXTENSION_CATEGORIES } from '../../../../platform/extensions/common/extensions.js';
import { productSchemaId } from '../../../../platform/product/common/productService.js';
import { ImplicitActivationEvents } from '../../../../platform/extensionManagement/common/implicitActivationEvents.js';
import { allApiProposals } from '../../../../platform/extensions/common/extensionsApiProposals.js';

const schemaRegistry = ( Registry.as(Extensions.JSONContribution));
class ExtensionMessageCollector {
    constructor(messageHandler, extension, extensionPointId) {
        this._messageHandler = messageHandler;
        this._extension = extension;
        this._extensionPointId = extensionPointId;
    }
    _msg(type, message) {
        this._messageHandler({
            type: type,
            message: message,
            extensionId: this._extension.identifier,
            extensionPointId: this._extensionPointId
        });
    }
    error(message) {
        this._msg(Severity$1.Error, message);
    }
    warn(message) {
        this._msg(Severity$1.Warning, message);
    }
    info(message) {
        this._msg(Severity$1.Info, message);
    }
}
class ExtensionPointUserDelta {
    static _toSet(arr) {
        const result = ( new ExtensionIdentifierSet());
        for (let i = 0, len = arr.length; i < len; i++) {
            result.add(arr[i].description.identifier);
        }
        return result;
    }
    static compute(previous, current) {
        if (!previous || !previous.length) {
            return ( new ExtensionPointUserDelta(current, []));
        }
        if (!current || !current.length) {
            return ( new ExtensionPointUserDelta([], previous));
        }
        const previousSet = this._toSet(previous);
        const currentSet = this._toSet(current);
        const added = current.filter(user => !( previousSet.has(user.description.identifier)));
        const removed = previous.filter(user => !( currentSet.has(user.description.identifier)));
        return ( new ExtensionPointUserDelta(added, removed));
    }
    constructor(added, removed) {
        this.added = added;
        this.removed = removed;
    }
}
class ExtensionPoint {
    constructor(name, defaultExtensionKind, canHandleResolver) {
        this.name = name;
        this.defaultExtensionKind = defaultExtensionKind;
        this.canHandleResolver = canHandleResolver;
        this._handler = null;
        this._users = null;
        this._delta = null;
    }
    setHandler(handler) {
        if (this._handler !== null) {
            throw ( new Error('Handler already set!'));
        }
        this._handler = handler;
        this._handle();
        return {
            dispose: () => {
                this._handler = null;
            }
        };
    }
    acceptUsers(users) {
        this._delta = ExtensionPointUserDelta.compute(this._users, users);
        this._users = users;
        this._handle();
    }
    _handle() {
        if (this._handler === null || this._users === null || this._delta === null) {
            return;
        }
        try {
            this._handler(this._users, this._delta);
        }
        catch (err) {
            onUnexpectedError(err);
        }
    }
}
const extensionKindSchema = {
    type: 'string',
    enum: [
        'ui',
        'workspace'
    ],
    enumDescriptions: [
        ( localize(
        12603,
        "UI extension kind. In a remote window, such extensions are enabled only when available on the local machine."
    )),
        ( localize(
        12604,
        "Workspace extension kind. In a remote window, such extensions are enabled only when available on the remote."
    )),
    ],
};
const schemaId = 'vscode://schemas/vscode-extensions';
const schema = {
    properties: {
        engines: {
            type: 'object',
            description: ( localize(12605, "Engine compatibility.")),
            properties: {
                'vscode': {
                    type: 'string',
                    description: ( localize(
                        12606,
                        'For VS Code extensions, specifies the VS Code version that the extension is compatible with. Cannot be *. For example: ^0.10.5 indicates compatibility with a minimum VS Code version of 0.10.5.'
                    )),
                    default: '^1.22.0',
                }
            }
        },
        publisher: {
            description: ( localize(12607, 'The publisher of the VS Code extension.')),
            type: 'string'
        },
        displayName: {
            description: ( localize(12608, 'The display name for the extension used in the VS Code gallery.')),
            type: 'string'
        },
        categories: {
            description: ( localize(
                12609,
                'The categories used by the VS Code gallery to categorize the extension.'
            )),
            type: 'array',
            uniqueItems: true,
            items: {
                oneOf: [{
                        type: 'string',
                        enum: EXTENSION_CATEGORIES,
                    },
                    {
                        type: 'string',
                        const: 'Languages',
                        deprecationMessage: ( localize(12610, 'Use \'Programming  Languages\' instead')),
                    }]
            }
        },
        galleryBanner: {
            type: 'object',
            description: ( localize(12611, 'Banner used in the VS Code marketplace.')),
            properties: {
                color: {
                    description: ( localize(12612, 'The banner color on the VS Code marketplace page header.')),
                    type: 'string'
                },
                theme: {
                    description: ( localize(12613, 'The color theme for the font used in the banner.')),
                    type: 'string',
                    enum: ['dark', 'light']
                }
            }
        },
        contributes: {
            description: ( localize(
                12614,
                'All contributions of the VS Code extension represented by this package.'
            )),
            type: 'object',
            properties: {
            },
            default: {}
        },
        preview: {
            type: 'boolean',
            description: ( localize(12615, 'Sets the extension to be flagged as a Preview in the Marketplace.')),
        },
        enableProposedApi: {
            type: 'boolean',
            deprecationMessage: ( localize(12616, 'Use `enabledApiProposals` instead.')),
        },
        enabledApiProposals: {
            markdownDescription: ( localize(
                12617,
                'Enable API proposals to try them out. Only valid **during development**. Extensions **cannot be published** with this property. For more details visit: https://code.visualstudio.com/api/advanced-topics/using-proposed-api'
            )),
            type: 'array',
            uniqueItems: true,
            items: {
                type: 'string',
                enum: ( ( Object.keys(allApiProposals)).map(proposalName => proposalName)),
                markdownEnumDescriptions: ( ( Object.values(allApiProposals)).map(value => value.proposal))
            }
        },
        api: {
            markdownDescription: ( localize(
                12618,
                'Describe the API provided by this extension. For more details visit: https://code.visualstudio.com/api/advanced-topics/remote-extensions#handling-dependencies-with-remote-extensions'
            )),
            type: 'string',
            enum: ['none'],
            enumDescriptions: [
                ( localize(
                12619,
                "Give up entirely the ability to export any APIs. This allows other extensions that depend on this extension to run in a separate extension host process or in a remote machine."
            ))
            ]
        },
        activationEvents: {
            description: ( localize(12620, 'Activation events for the VS Code extension.')),
            type: 'array',
            items: {
                type: 'string',
                defaultSnippets: [
                    {
                        label: 'onWebviewPanel',
                        description: ( localize(
                            12621,
                            'An activation event emmited when a webview is loaded of a certain viewType'
                        )),
                        body: 'onWebviewPanel:viewType'
                    },
                    {
                        label: 'onLanguage',
                        description: ( localize(
                            12622,
                            'An activation event emitted whenever a file that resolves to the specified language gets opened.'
                        )),
                        body: 'onLanguage:${1:languageId}'
                    },
                    {
                        label: 'onCommand',
                        description: ( localize(
                            12623,
                            'An activation event emitted whenever the specified command gets invoked.'
                        )),
                        body: 'onCommand:${2:commandId}'
                    },
                    {
                        label: 'onDebug',
                        description: ( localize(
                            12624,
                            'An activation event emitted whenever a user is about to start debugging or about to setup debug configurations.'
                        )),
                        body: 'onDebug'
                    },
                    {
                        label: 'onDebugInitialConfigurations',
                        description: ( localize(
                            12625,
                            'An activation event emitted whenever a "launch.json" needs to be created (and all provideDebugConfigurations methods need to be called).'
                        )),
                        body: 'onDebugInitialConfigurations'
                    },
                    {
                        label: 'onDebugDynamicConfigurations',
                        description: ( localize(
                            12626,
                            'An activation event emitted whenever a list of all debug configurations needs to be created (and all provideDebugConfigurations methods for the "dynamic" scope need to be called).'
                        )),
                        body: 'onDebugDynamicConfigurations'
                    },
                    {
                        label: 'onDebugResolve',
                        description: ( localize(
                            12627,
                            'An activation event emitted whenever a debug session with the specific type is about to be launched (and a corresponding resolveDebugConfiguration method needs to be called).'
                        )),
                        body: 'onDebugResolve:${6:type}'
                    },
                    {
                        label: 'onDebugAdapterProtocolTracker',
                        description: ( localize(
                            12628,
                            'An activation event emitted whenever a debug session with the specific type is about to be launched and a debug protocol tracker might be needed.'
                        )),
                        body: 'onDebugAdapterProtocolTracker:${6:type}'
                    },
                    {
                        label: 'workspaceContains',
                        description: ( localize(
                            12629,
                            'An activation event emitted whenever a folder is opened that contains at least a file matching the specified glob pattern.'
                        )),
                        body: 'workspaceContains:${4:filePattern}'
                    },
                    {
                        label: 'onStartupFinished',
                        description: ( localize(
                            12630,
                            'An activation event emitted after the start-up finished (after all `*` activated extensions have finished activating).'
                        )),
                        body: 'onStartupFinished'
                    },
                    {
                        label: 'onTaskType',
                        description: ( localize(
                            12631,
                            'An activation event emitted whenever tasks of a certain type need to be listed or resolved.'
                        )),
                        body: 'onTaskType:${1:taskType}'
                    },
                    {
                        label: 'onFileSystem',
                        description: ( localize(
                            12632,
                            'An activation event emitted whenever a file or folder is accessed with the given scheme.'
                        )),
                        body: 'onFileSystem:${1:scheme}'
                    },
                    {
                        label: 'onEditSession',
                        description: ( localize(
                            12633,
                            'An activation event emitted whenever an edit session is accessed with the given scheme.'
                        )),
                        body: 'onEditSession:${1:scheme}'
                    },
                    {
                        label: 'onSearch',
                        description: ( localize(
                            12634,
                            'An activation event emitted whenever a search is started in the folder with the given scheme.'
                        )),
                        body: 'onSearch:${7:scheme}'
                    },
                    {
                        label: 'onView',
                        body: 'onView:${5:viewId}',
                        description: ( localize(
                            12635,
                            'An activation event emitted whenever the specified view is expanded.'
                        )),
                    },
                    {
                        label: 'onUri',
                        body: 'onUri',
                        description: ( localize(
                            12636,
                            'An activation event emitted whenever a system-wide Uri directed towards this extension is open.'
                        )),
                    },
                    {
                        label: 'onOpenExternalUri',
                        body: 'onOpenExternalUri',
                        description: ( localize(
                            12637,
                            'An activation event emitted whenever a external uri (such as an http or https link) is being opened.'
                        )),
                    },
                    {
                        label: 'onCustomEditor',
                        body: 'onCustomEditor:${9:viewType}',
                        description: ( localize(
                            12638,
                            'An activation event emitted whenever the specified custom editor becomes visible.'
                        )),
                    },
                    {
                        label: 'onNotebook',
                        body: 'onNotebook:${1:type}',
                        description: ( localize(
                            12639,
                            'An activation event emitted whenever the specified notebook document is opened.'
                        )),
                    },
                    {
                        label: 'onAuthenticationRequest',
                        body: 'onAuthenticationRequest:${11:authenticationProviderId}',
                        description: ( localize(
                            12640,
                            'An activation event emitted whenever sessions are requested from the specified authentication provider.'
                        ))
                    },
                    {
                        label: 'onRenderer',
                        description: ( localize(
                            12641,
                            'An activation event emitted whenever a notebook output renderer is used.'
                        )),
                        body: 'onRenderer:${11:rendererId}'
                    },
                    {
                        label: 'onTerminalProfile',
                        body: 'onTerminalProfile:${1:terminalId}',
                        description: ( localize(
                            12642,
                            'An activation event emitted when a specific terminal profile is launched.'
                        )),
                    },
                    {
                        label: 'onTerminalQuickFixRequest',
                        body: 'onTerminalQuickFixRequest:${1:quickFixId}',
                        description: ( localize(
                            12643,
                            'An activation event emitted when a command matches the selector associated with this ID'
                        )),
                    },
                    {
                        label: 'onWalkthrough',
                        body: 'onWalkthrough:${1:walkthroughID}',
                        description: ( localize(
                            12644,
                            'An activation event emitted when a specified walkthrough is opened.'
                        )),
                    },
                    {
                        label: 'onIssueReporterOpened',
                        body: 'onIssueReporterOpened',
                        description: ( localize(12645, 'An activation event emitted when the issue reporter is opened.')),
                    },
                    {
                        label: 'onChatParticipant',
                        body: 'onChatParticipant:${1:participantId}',
                        description: ( localize(
                            12646,
                            'An activation event emitted when the specified chat participant is invoked.'
                        )),
                    },
                    {
                        label: 'onLanguageModelTool',
                        body: 'onLanguageModelTool:${1:toolId}',
                        description: ( localize(
                            12647,
                            'An activation event emitted when the specified language model tool is invoked.'
                        )),
                    },
                    {
                        label: 'onTerminalCompletionsRequested',
                        body: 'onTerminalCompletionsRequested',
                        description: ( localize(
                            12648,
                            'An activation event emitted when terminal completions are requested.'
                        )),
                    },
                    {
                        label: 'onMcpCollection',
                        description: ( localize(
                            12649,
                            'An activation event emitted whenver a tool from the MCP server is requested.'
                        )),
                        body: 'onMcpCollection:${2:collectionId}',
                    },
                    {
                        label: '*',
                        description: ( localize(
                            12650,
                            'An activation event emitted on VS Code startup. To ensure a great end user experience, please use this activation event in your extension only when no other activation events combination works in your use-case.'
                        )),
                        body: '*'
                    }
                ],
            }
        },
        badges: {
            type: 'array',
            description: ( localize(
                12651,
                'Array of badges to display in the sidebar of the Marketplace\'s extension page.'
            )),
            items: {
                type: 'object',
                required: ['url', 'href', 'description'],
                properties: {
                    url: {
                        type: 'string',
                        description: ( localize(12652, 'Badge image URL.'))
                    },
                    href: {
                        type: 'string',
                        description: ( localize(12653, 'Badge link.'))
                    },
                    description: {
                        type: 'string',
                        description: ( localize(12654, 'Badge description.'))
                    }
                }
            }
        },
        markdown: {
            type: 'string',
            description: ( localize(
                12655,
                "Controls the Markdown rendering engine used in the Marketplace. Either github (default) or standard."
            )),
            enum: ['github', 'standard'],
            default: 'github'
        },
        qna: {
            default: 'marketplace',
            description: ( localize(
                12656,
                "Controls the Q&A link in the Marketplace. Set to marketplace to enable the default Marketplace Q & A site. Set to a string to provide the URL of a custom Q & A site. Set to false to disable Q & A altogether."
            )),
            anyOf: [
                {
                    type: ['string', 'boolean'],
                    enum: ['marketplace', false]
                },
                {
                    type: 'string'
                }
            ]
        },
        extensionDependencies: {
            description: ( localize(
                12657,
                'Dependencies to other extensions. The identifier of an extension is always ${publisher}.${name}. For example: vscode.csharp.'
            )),
            type: 'array',
            uniqueItems: true,
            items: {
                type: 'string',
                pattern: EXTENSION_IDENTIFIER_PATTERN
            }
        },
        extensionPack: {
            description: ( localize(
                12658,
                "A set of extensions that can be installed together. The identifier of an extension is always ${publisher}.${name}. For example: vscode.csharp."
            )),
            type: 'array',
            uniqueItems: true,
            items: {
                type: 'string',
                pattern: EXTENSION_IDENTIFIER_PATTERN
            }
        },
        extensionKind: {
            description: ( localize(
                12659,
                "Define the kind of an extension. `ui` extensions are installed and run on the local machine while `workspace` extensions run on the remote."
            )),
            type: 'array',
            items: extensionKindSchema,
            default: ['workspace'],
            defaultSnippets: [
                {
                    body: ['ui'],
                    description: ( localize(
                        12660,
                        "Define an extension which can run only on the local machine when connected to remote window."
                    ))
                },
                {
                    body: ['workspace'],
                    description: ( localize(
                        12661,
                        "Define an extension which can run only on the remote machine when connected remote window."
                    ))
                },
                {
                    body: ['ui', 'workspace'],
                    description: ( localize(
                        12662,
                        "Define an extension which can run on either side, with a preference towards running on the local machine."
                    ))
                },
                {
                    body: ['workspace', 'ui'],
                    description: ( localize(
                        12663,
                        "Define an extension which can run on either side, with a preference towards running on the remote machine."
                    ))
                },
                {
                    body: [],
                    description: ( localize(
                        12664,
                        "Define an extension which cannot run in a remote context, neither on the local, nor on the remote machine."
                    ))
                }
            ]
        },
        capabilities: {
            description: ( localize(12665, "Declare the set of supported capabilities by the extension.")),
            type: 'object',
            properties: {
                virtualWorkspaces: {
                    description: ( localize(
                        12666,
                        "Declares whether the extension should be enabled in virtual workspaces. A virtual workspace is a workspace which is not backed by any on-disk resources. When false, this extension will be automatically disabled in virtual workspaces. Default is true."
                    )),
                    type: ['boolean', 'object'],
                    defaultSnippets: [
                        { label: 'limited', body: { supported: '${1:limited}', description: '${2}' } },
                        { label: 'false', body: { supported: false, description: '${2}' } },
                    ],
                    default: true.valueOf,
                    properties: {
                        supported: {
                            markdownDescription: ( localize(
                                12667,
                                "Declares the level of support for virtual workspaces by the extension."
                            )),
                            type: ['string', 'boolean'],
                            enum: ['limited', true, false],
                            enumDescriptions: [
                                ( localize(
                                12668,
                                "The extension will be enabled in virtual workspaces with some functionality disabled."
                            )),
                                ( localize(
                                12669,
                                "The extension will be enabled in virtual workspaces with all functionality enabled."
                            )),
                                ( localize(12670, "The extension will not be enabled in virtual workspaces.")),
                            ]
                        },
                        description: {
                            type: 'string',
                            markdownDescription: ( localize(
                                12671,
                                "A description of how virtual workspaces affects the extensions behavior and why it is needed. This only applies when `supported` is not `true`."
                            )),
                        }
                    }
                },
                untrustedWorkspaces: {
                    description: ( localize(
                        12672,
                        'Declares how the extension should be handled in untrusted workspaces.'
                    )),
                    type: 'object',
                    required: ['supported'],
                    defaultSnippets: [
                        { body: { supported: '${1:limited}', description: '${2}' } },
                    ],
                    properties: {
                        supported: {
                            markdownDescription: ( localize(
                                12673,
                                "Declares the level of support for untrusted workspaces by the extension."
                            )),
                            type: ['string', 'boolean'],
                            enum: ['limited', true, false],
                            enumDescriptions: [
                                ( localize(
                                12674,
                                "The extension will be enabled in untrusted workspaces with some functionality disabled."
                            )),
                                ( localize(
                                12675,
                                "The extension will be enabled in untrusted workspaces with all functionality enabled."
                            )),
                                ( localize(12676, "The extension will not be enabled in untrusted workspaces.")),
                            ]
                        },
                        restrictedConfigurations: {
                            description: ( localize(
                                12677,
                                "A list of configuration keys contributed by the extension that should not use workspace values in untrusted workspaces."
                            )),
                            type: 'array',
                            items: {
                                type: 'string'
                            }
                        },
                        description: {
                            type: 'string',
                            markdownDescription: ( localize(
                                12678,
                                "A description of how workspace trust affects the extensions behavior and why it is needed. This only applies when `supported` is not `true`."
                            )),
                        }
                    }
                }
            }
        },
        sponsor: {
            description: ( localize(12679, "Specify the location from where users can sponsor your extension.")),
            type: 'object',
            defaultSnippets: [
                { body: { url: '${1:https:}' } },
            ],
            properties: {
                'url': {
                    description: ( localize(
                        12680,
                        "URL from where users can sponsor your extension. It must be a valid URL with a HTTP or HTTPS protocol. Example value: https://github.com/sponsors/nvaccess"
                    )),
                    type: 'string',
                }
            }
        },
        scripts: {
            type: 'object',
            properties: {
                'vscode:prepublish': {
                    description: ( localize(
                        12681,
                        'Script executed before the package is published as a VS Code extension.'
                    )),
                    type: 'string'
                },
                'vscode:uninstall': {
                    description: ( localize(
                        12682,
                        'Uninstall hook for VS Code extension. Script that gets executed when the extension is completely uninstalled from VS Code which is when VS Code is restarted (shutdown and start) after the extension is uninstalled. Only Node scripts are supported.'
                    )),
                    type: 'string'
                }
            }
        },
        icon: {
            type: 'string',
            description: ( localize(12683, 'The path to a 128x128 pixel icon.'))
        },
        l10n: {
            type: 'string',
            description: ( localize(
                12684,
                'The relative path to a folder containing localization (bundle.l10n.*.json) files. Must be specified if you are using the vscode.l10n API.'
            ))
        },
        pricing: {
            type: 'string',
            markdownDescription: ( localize(
                12685,
                'The pricing information for the extension. Can be Free (default) or Trial. For more details visit: https://code.visualstudio.com/api/working-with-extensions/publishing-extension#extension-pricing-label'
            )),
            enum: ['Free', 'Trial'],
            default: 'Free'
        }
    }
};
class ExtensionsRegistryImpl {
    constructor() {
        this._extensionPoints = ( new Map());
    }
    registerExtensionPoint(desc) {
        if (( this._extensionPoints.has(desc.extensionPoint))) {
            throw ( new Error('Duplicate extension point: ' + desc.extensionPoint));
        }
        const result = ( new ExtensionPoint(desc.extensionPoint, desc.defaultExtensionKind, desc.canHandleResolver));
        this._extensionPoints.set(desc.extensionPoint, result);
        if (desc.activationEventsGenerator) {
            ImplicitActivationEvents.register(desc.extensionPoint, desc.activationEventsGenerator);
        }
        schema.properties['contributes'].properties[desc.extensionPoint] = desc.jsonSchema;
        schemaRegistry.registerSchema(schemaId, schema);
        return result;
    }
    getExtensionPoints() {
        return Array.from(( this._extensionPoints.values()));
    }
}
const PRExtensions = {
    ExtensionsRegistry: 'ExtensionsRegistry'
};
Registry.add(PRExtensions.ExtensionsRegistry, ( new ExtensionsRegistryImpl()));
const ExtensionsRegistry = ( Registry.as(PRExtensions.ExtensionsRegistry));
schemaRegistry.registerSchema(schemaId, schema);
schemaRegistry.registerSchema(productSchemaId, {
    properties: {
        extensionEnabledApiProposals: {
            description: ( localize(12686, "API proposals that the respective extensions can freely use.")),
            type: 'object',
            properties: {},
            additionalProperties: {
                anyOf: [{
                        type: 'array',
                        uniqueItems: true,
                        items: {
                            type: 'string',
                            enum: ( Object.keys(allApiProposals)),
                            markdownEnumDescriptions: ( ( Object.values(allApiProposals)).map(value => value.proposal))
                        }
                    }]
            }
        }
    }
});

export { ExtensionMessageCollector, ExtensionPoint, ExtensionPointUserDelta, ExtensionsRegistry, ExtensionsRegistryImpl, schema };
